/*
    Title:    DCF77 library
    Author:   Volker Oth
              based on the terminal example by 
              Robert Mashlan (rmashlan@r2m.com)
    Date:     5/1999
    Purpose:  Decoding of the DCF77 real time signal.
    needed
    Software: MS VCC 5.0
    needed
    Hardware: ATS90S8515 on STK200 board delivering dcf77 signal
              via uart
    Note:     To contact me, mail to
                  volkeroth@gmx.de
              You might find more AVR related stuff at my homepage:
                  http://members.xoom.com/volkeroth
*/
#define STRICT
#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

typedef unsigned char  u08;
typedef signed   char  s08;
typedef unsigned short u16;
typedef signed   short s16;
typedef unsigned long  u32;
typedef signed   long  s32;

#ifndef LANGUAGE         /* 0: German, 1:English */
#define LANGUAGE 0
#endif

#define DCF77_STATUS_BIT         0         /* last bit received */
#define DCF77_STATUS_INIT        1         /* flag: start detected */
#define DCF77_STATUS_MIN_VALID   2         /* flag: minute info valid */
#define DCF77_STATUS_HOUR_VALID  3         /* flag: hour info valid */
#define DCF77_STATUS_DATE_VALID  4         /* flag: day info valid */

#define DCF77_BUFFER_SIZE        9

#define dcf77_start  dcf77_data[0]         /* start ID */ 
#define dcf77_second dcf77_data[1]         /* current second */
#define dcf77_minute dcf77_data[2]         /* decoded minute */ 
#define dcf77_hour   dcf77_data[3]         /* decoded hour */
#define dcf77_day    dcf77_data[4]         /* decoded day */
#define dcf77_wday   dcf77_data[5]         /* decoded weekday */
#define dcf77_month  dcf77_data[6]         /* decoded month */ 
#define dcf77_year   dcf77_data[7]         /* decoded year */ 
#define dcf77_status dcf77_data[8]         /* status bit field */ 

u08 print_format;
u08 dcf77_data[DCF77_BUFFER_SIZE];
#if LANGUAGE==0
char dayname[8][3] = { "??", "Mo", "Di", "Mi", "Do", "Fr", "Sa" , "So" };
#else 
char dayname[8][3] = { "??", "Mo", "Tu", "We", "Th", "Fr", "Sa" , "Su" };
#endif


void print_time(void)
{
    register u08 i;
    char str_day[3];
    char str_wday[3];
    char str_month[3];
    char str_year[3];
    char str_min[3];
    char str_hour[3];

 
    switch (print_format) {
        case 0: /* print bits */
            if ((dcf77_second)==0) printf("\n");
            if (dcf77_status & (1<<DCF77_STATUS_BIT)) 
                printf("*");
            else           
                printf("-");
            break;
        case 1: /* print raw data */
            for (i=0; i< DCF77_BUFFER_SIZE; i++) 
                printf("%02X ",dcf77_data[i]);
            puts("");
            break;
        case 2: /* print formatted time */
            if (!(dcf77_status & (1<<DCF77_STATUS_INIT)))
#if LANGUAGE==0
                printf("synchronisieren %02u\n", dcf77_second);
#else
                printf("synchronizing %02u\n",   dcf77_second);
#endif
            else {
                if (dcf77_status & (1<<DCF77_STATUS_DATE_VALID)) {
                    sprintf( str_day  , "%02u", dcf77_day   );
                    strcpy(  str_wday , dayname[dcf77_wday] );
                    sprintf( str_month, "%02u", dcf77_month );
                    sprintf( str_year,  "%02u", dcf77_year  );
                }
                else {
                    strcpy( str_day,   "??" );
                    strcpy( str_wday,  "??" );
                    strcpy( str_month, "??" );
                    strcpy( str_year,  "??" );
                }
                if (dcf77_status & (1<<DCF77_STATUS_MIN_VALID)) 
                    sprintf( str_min  , "%02u", dcf77_minute);
                else 
                    strcpy( str_min,   "??" );
                if (dcf77_status & (1<<DCF77_STATUS_HOUR_VALID)) 
                    sprintf( str_hour , "%02u", dcf77_hour  );
                else 
                    strcpy( str_hour,  "??" );

#if LANGUAGE==0
                printf("%2s.%2s.%2s (%2s) %2s:%2s:%02u\n",
                    str_day, str_month, str_year, str_wday, 
                    str_hour, str_min, dcf77_second);
#else
                printf("%2s/%2s/%2s (%2s) %2s:%2s:%02u\n",
                    str_month, str_day, str_year, str_wday, 
                    str_hour, str_min, dcf77_second);
#endif
            }
    }
}


void read_uart( HANDLE h )
{
    HANDLE hconn = GetStdHandle(STD_OUTPUT_HANDLE);
    DWORD mask;
    OVERLAPPED ov;
    static read_bytes;

    ZeroMemory(&ov,sizeof(ov));

    // create event for overlapped I/O
    ov.hEvent = CreateEvent(NULL,FALSE,FALSE,NULL);
    if(ov.hEvent == INVALID_HANDLE_VALUE)
        puts("CreateEvent failed");

    // wait for received characters 
    if(!SetCommMask(h,EV_RXCHAR))
        puts("SetCommMask failed");

    while( 1 ) {
        // get the event mask
        if( !WaitCommEvent(h,&mask,&ov) ) {
            DWORD e = GetLastError();
            if( e == ERROR_IO_PENDING ) {
                DWORD r;
                if( !GetOverlappedResult(h,&ov,&r,TRUE) ) {
                    puts("GetOverlappedResult failed");
                    break;
                }
            } 
            else {
                puts("WaitCommEvent failed");
                break;
            }
        }

        // if no event, then UI thread terminated with SetCommMask(h,0)
        if( mask == 0 ) 
            break;

        if( mask & EV_RXCHAR) {
            u08 buf[16];
            DWORD read;
            register u08 i;
            do {
                read = 0;
                if( !ReadFile(h,buf,sizeof(buf),&read,&ov) ) {
                    if( GetLastError() == ERROR_IO_PENDING ) {
                        if( !GetOverlappedResult(h,&ov,&read,TRUE) ) {
                            puts("GetOverlappedResult failed");
                            break;
                        }
                    } 
                    else {
                        puts("ReadFile failed");
                        break;
                    }
                }

                for (i=0; i<read; i++) {
                    if (buf[i] == (u08)0xfe) {
                        dcf77_data[0] = 0xfe;
                        read_bytes = 1;
                    }
                    else if (read_bytes) {
                        dcf77_data[read_bytes] = buf[i];
                        read_bytes++;
                    }
                    if (read_bytes == DCF77_BUFFER_SIZE) {
                        print_time();
                        read_bytes = 0;
                    }                 
                }

                /*WriteFile(hconn,buf,read,&read,NULL);*/
            } while(read);
        }
        mask = 0;
    }
    CloseHandle(ov.hEvent);  // close the event
}


int main( int argc, char **argv )
{
    HANDLE h;
    COMMTIMEOUTS cto = { 2, 1, 1, 0, 0 };             // set timeouts
    DCB dcb;

    if(argc>1) {
        if(argc>2)
            print_format = atoi(argv[2]);
        else
            print_format = 2;

        // open port for overlapped I/O
        h = CreateFile(argv[1], GENERIC_READ|GENERIC_WRITE, 0,NULL,
                       OPEN_EXISTING,FILE_FLAG_OVERLAPPED,NULL);

        if (h == INVALID_HANDLE_VALUE) 
            puts("Failed to open port"); 
        else {
            if(!SetCommTimeouts(h,&cto))
                puts("SetCommTimeouts failed");

            // set DCB
            memset(&dcb,0,sizeof(dcb));
            dcb.DCBlength = sizeof(dcb);
            dcb.BaudRate = CBR_9600;
            dcb.fBinary = 1;
            dcb.Parity = NOPARITY;
            dcb.fOutxCtsFlow = 0;
            dcb.fOutxDsrFlow = 0;
            dcb.fDsrSensitivity = 0;

            dcb.fDtrControl = DTR_CONTROL_DISABLE;
            dcb.fRtsControl = RTS_CONTROL_DISABLE;

            dcb.StopBits = ONESTOPBIT;
            dcb.ByteSize = 8;

            if(!SetCommState(h,&dcb))
                puts("SetCommState failed");

            // start terminal
            read_uart(h);

            CloseHandle(h);
        }
    } 
    else {
        puts("dcf77 port [format]");
        puts("  formats:");
        puts("  0: 60 bit output");
        puts("  1: raw byte output");
        puts("  2: formatted output [default]");
    }
    
   return 0;
}